!
! Program to plot disconnectivity trees.
! See Becker and Karplus JCP 106 1495 (1997).
!
! Information and options are passed to the program in a file called
! "dinfo", which is keyword driven.
!
! Compulsory Keywords
! -------------------
!
! DELTA <dE>
! Energetic separation of levels in basin analysis.
!
! FIRST <E1>
! Specifies the energy of the highest level on the energy axis.
!
! LEVELS <n>
! The number of levels at which to perform the basin analysis.
!
! MINIMA <file>
! Specifies filename for minima info.
!
! TS <file>
! Specifies filename for transition state info.
!
!
! Optional Keywords
! -----------------
!
! CENTREGMIN
! If this keyword is present, then when a node splits into its daughter
! nodes, the one containing the global minimum is always placed centrally
! (even if other nodes carry more minima). This does not guarantee that
! the global minimum is central in the overall diagram because other
! nodes may push the one containing the global minimum over to one side.
!
! DUMPNUMBERS
! If present, a file called node_numbers is written, listing the minima
! associated with each node in each level. Nodes are listed from left to
! right within each level.
!
! DUMPSIZES
! If present, a file called node_sizes is written, listing how many minima
! are represented by each node in each level. Nodes are listed from left to
! right in each level.
!
! EXCLUDEALL
! Removes all minima from the list of minima to be plotted.  This is to be
! used in conjunction with the PICK command which can be used to specify
! exclusively which minima are to be included.
!
! IDENTIFY
! If present, the branch ends are labelled with the lowest-energy minimum
! they represent.
!
! LABELFORMAT <fmt>
! Specifies the Fortran format string for the energt level labels. The default
! is F6.1.
!
! LETTER
! If present, the graph is formatted for American letter paper rather than
! European A4.
!
! LOWEST <n>
! If present, only the branches leading to the lowest n minima are drawn. The
! pruning occurs after the basin analysis, so the "discarded" minima can still
! influence the connectivities.
!
! MONOTONIC
! If present, all minima not lying at the bottom of a monotonic sequences are
! not drawn. This tends to reduce the number of branches drastically. If the
! keyword LOWEST is also used, the MONOTONIC sequence analysis is applied after
! the high energy minima have been discarded.
!
! NOBARRIERS
! If present, all transition state energies are reset to the energy of the higher
! of the two minima they connect. This transforms the energy landscape to the
! type explored by gmin.
!
! PICK <file>
! Specifies the name of a list of numbers of minima, one per line.  Minima on
! this list are included on the graph.  Minima preceded with a minus sign are
! removed from the graph.  This process is executed after the commands
! MONOTONIC, LOWEST and EXCLUDEALL have been executed, thereby making it
! possible to override them for particular minima.  Examples: 1. To remove
! certain minima from a full plot, just specify PICK and a list of negative
! minima numbers.  2. To include only specific minima, use EXCLUDEALL and
! PICK plus a list of positive minima numbers.  All basin analysis includes
! the full sample and is performed before minima are removed or added back in.
!
! NOSPLIT
! By default, every minimum is indicated by its own branch, which splits off
! from the parent basin even if the minimum and its lowest transition state do
! not straddle an energy level in the basin analysis. This is to avoid it being
! dependent on precisely where the levels are placed (bulk shifting of the levels
! would cause some branch ends to appear or disappear rather than change node
! if this were not the case). The NOSPLIT option turns this feature off, so that
! if two minima are separated by a barrier lower than the level above their own
! energy, they are grouped together. This option should probably never be used.

!................................................................................!

!!!
!!! KEYWORDS LIBRARY
!!!
!!! Subroutines: read_line(unit[,logical])
!!!              get_string(string[,logical])
!!!              get_integer(integer[,logical])
!!!              get_dp(dp[,logical])
!!!              get_logical(lgcl[,logical])
!!!              upper_case(string)
!!!
!!! Version 2.0
!!! MM 21.ix.96
!!!

MODULE keywords
   IMPLICIT NONE
   INTEGER, PARAMETER :: max_lines=2, max_length=100
   INTEGER, PARAMETER :: tot_length=max_lines*max_length
   INTEGER :: position
   CHARACTER(LEN=tot_length) :: input
   SAVE
   CONTAINS
!!
   SUBROUTINE read_line(u, success)
      IMPLICIT NONE
      INTEGER, INTENT(IN) :: u
      LOGICAL, INTENT(INOUT), OPTIONAL :: success
      INTEGER :: i, lines, next, err
      LOGICAL :: continue
      CHARACTER(LEN=7) :: fmt_string
!     Generate format string of max_length characters.
      WRITE (fmt_string, '(I4)') max_length
      fmt_string = '(A' // TRIM(ADJUSTL(fmt_string)) // ')'
      DO
         input=' '
         next = 1
!        Read in a logical line consisting of up to max_lines of input file.
         DO lines=1, max_lines
            continue = .FALSE.
            READ (u, fmt_string, IOSTAT=err) input(next:next+max_length-1)
            IF (err == 0) THEN
               IF (PRESENT(success)) success = .TRUE.
            ELSE
               IF (PRESENT(success)) success = .FALSE.
               EXIT
            ENDIF
!           Check for continuation symbol (&).
            DO i=next, next+max_length-1
               IF (input(i:i)=='&') THEN
                  continue = .TRUE.
                  next = i
                  EXIT
               ENDIF
            END DO
            IF (.NOT.continue) EXIT
         END DO
         IF (err /= 0) EXIT
         IF (TRIM(input) /= '') EXIT   ! Only read in next line if this one's empty.
      END DO
      position = 1
   END SUBROUTINE read_line
!!
   SUBROUTINE upper_case(string)
      IMPLICIT NONE
      CHARACTER(LEN=*), INTENT(INOUT) :: string
      INTEGER, PARAMETER :: lower_to_upper = ICHAR("A")-ICHAR("a")
      INTEGER :: i
      DO i=1, LEN_TRIM(string)
         IF (LGE(string(i:i), 'a').AND.LLE(string(i:i), 'z')) THEN
            string(i:i) = ACHAR(IACHAR(string(i:i))+lower_to_upper)
         ENDIF
      END DO
   END SUBROUTINE upper_case
!!
   SUBROUTINE get_string(string, success)
      IMPLICIT NONE
      CHARACTER(LEN=*), INTENT(INOUT) :: string
      LOGICAL, INTENT(INOUT), OPTIONAL :: success
      CHARACTER(LEN=tot_length) :: temp
      INTEGER :: outcome
      CALL next_item(temp, outcome)
      IF (outcome == 3) THEN
         string = temp
         IF (PRESENT(success)) success = .TRUE.
      ELSE
         IF (PRESENT(success)) success = .FALSE.
      ENDIF
   END SUBROUTINE get_string
!!
   SUBROUTINE get_integer(value, success)
      IMPLICIT NONE
      INTEGER, INTENT(INOUT) :: value
      LOGICAL, INTENT(INOUT), OPTIONAL :: success
      INTEGER :: temp, outcome, err
      CHARACTER(LEN=tot_length) :: item
      CALL next_item(item, outcome)
      READ (UNIT=item, FMT=*, IOSTAT=err) temp
      IF ((err == 0).AND.(outcome==3)) THEN
         value = temp
         IF (PRESENT(success)) success = .TRUE.
      ELSE
         IF (PRESENT(success)) success = .FALSE.
      ENDIF
   END SUBROUTINE get_integer
!!
   SUBROUTINE get_dp(value, success)
      IMPLICIT NONE
      DOUBLE PRECISION, INTENT(INOUT) :: value
      LOGICAL, INTENT(INOUT), OPTIONAL :: success
      DOUBLE PRECISION :: temp
      INTEGER :: outcome, err
      CHARACTER(LEN=tot_length) :: item
      CALL next_item(item, outcome)
      READ (UNIT=item, FMT=*, IOSTAT=err) temp
      IF ((err == 0).AND.(outcome==3)) THEN
         value = temp
         IF (PRESENT(success)) success = .TRUE.
      ELSE
         IF (PRESENT(success)) success = .FALSE.
      ENDIF
   END SUBROUTINE get_dp
!!
   SUBROUTINE get_logical(lgcl, success)
      IMPLICIT NONE
      LOGICAL, INTENT(INOUT) :: lgcl
      LOGICAL, INTENT(INOUT), OPTIONAL :: success
      INTEGER :: outcome
      CHARACTER(LEN=tot_length) :: item
      CALL next_item(item, outcome)
      CALL upper_case(item)
      IF ((TRIM(item)=='TRUE').OR.(TRIM(item)=='T').OR.(TRIM(item)=='.TRUE.') &
    & .OR.(TRIM(item)=='.T.').OR.(TRIM(item)=='ON')) THEN
         lgcl = .TRUE.
         IF (PRESENT(success)) success = .TRUE.
      ELSE IF ((TRIM(item)=='FALSE').OR.(TRIM(item)=='F').OR.(TRIM(item)=='.FALSE.') &
    & .OR.(TRIM(item)=='.F.').OR.(TRIM(item)=='OFF')) THEN
         lgcl = .FALSE.
         IF (PRESENT(success)) success = .TRUE.
      ELSE
         IF (PRESENT(success)) success = .FALSE.
      ENDIF
   END SUBROUTINE get_logical
!!
   SUBROUTINE next_item(item, outcome)
      IMPLICIT NONE
      CHARACTER(LEN=tot_length), INTENT(OUT) :: item
      INTEGER, INTENT(OUT) :: outcome
!     Values of outcome:
!      1: null string read
!      2: end of line reached with no string
!      3: correctly read a string of at least one character
      INTEGER :: i, j
      item = ''
      outcome=1
!     Check we've not already reached the end of the input string.
      IF (position > tot_length) THEN
         outcome = 2
      ELSE
!        Read past leading blanks.
         DO i=position, tot_length
            IF (input(i:i) /= ' ') EXIT
         END DO
!        Check that this hasn't brought us to the end of the input string.
         IF (i==tot_length+1) THEN
            outcome=2
         ELSE
            position = i
            j = 1
!           Read until the next space or comma.
            DO i=position, tot_length
               SELECT CASE(input(i:i))
!              If we've reached a comma, record the position for the next
!              item and exit loop.
               CASE(',')
                  position = i+1
                  EXIT
!              If we've reached a space, check for a comma preceded by some
!              blanks, and record the position for the next item as after the
!              comma if one is found.
               CASE(' ')
                  DO j=position+1, tot_length
                     SELECT CASE(input(j:j))
                     CASE(',')
                        position = j+1
                        EXIT
                     CASE (' ')   ! Do nothing.
                     CASE DEFAULT
                        position = j
                        EXIT
                     END SELECT
                  END DO
                  EXIT
!              Any other character is the next character of the item being read.
               CASE DEFAULT
                  item(j:j) = input(i:i)
                  j = j + 1
                  outcome=3
                  position = i+1
               END SELECT
            END DO
         ENDIF
      ENDIF
   END SUBROUTINE next_item
!!
END MODULE keywords

!................................................................................!

MODULE vars

   IMPLICIT NONE

   INTEGER, DIMENSION(:), ALLOCATABLE :: nodes
   INTEGER, DIMENSION(:,:), ALLOCATABLE :: basin, branches, col_0, order, &
      node_size, parent
   DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: end_x, end_y, m

   INTEGER :: n_levels, n_min, lowest
   DOUBLE PRECISION :: delta_e, e_high
   CHARACTER(LEN=120) :: file_min, file_pick, file_ts
   CHARACTER(LEN=10) :: lab_fmt
   LOGICAL :: barriers, centre_gmin, dump_numbers, dump_sizes, excludeall, &
      identify, monotonic, split

END MODULE vars

!................................................................................!

MODULE page

   IMPLICIT NONE
   INTEGER :: page_x=595,  page_y=842
   INTEGER :: margin_x=40, margin_y=50
   INTEGER :: scale_width=40
   INTEGER :: font_size=14
   INTEGER :: label_size=8

END MODULE page

!................................................................................!

PROGRAM disconnection

   USE page
   USE vars
   IMPLICIT NONE

   TYPE transition_state
      DOUBLE PRECISION :: e
      INTEGER :: n, min1, min2
   END TYPE transition_state

   CHARACTER(LEN=3) :: pg
   CHARACTER(LEN=8) :: min_trim
   DOUBLE PRECISION :: b_high, b_low, end_e, energy, frac, lpafs, &
      & x1, x2, x3, x_pos, y1, y2, y3, y_pos, padding, e_gmin, split_option
   TYPE(transition_state), DIMENSION(:), ALLOCATABLE :: ts
   INTEGER :: i, j, k, p, q, s
   INTEGER :: basin_no, end_m, err, h_pg, min1, min2, min_temp, n_br, n_cols, &
      & n_nodes, n_ts, path, used, big_ones, basin_gmin, gmin
   INTEGER, DIMENSION(:), ALLOCATABLE :: connect, indx1, sorted
   LOGICAL :: again

   WRITE (6, '(/, A)') 'Disconnectivity Graphs'
   WRITE (6, '(A, /)') '----------------------'

   CALL read_options


!  Procure minima info.
   CALL count_min(file_min, n_min)
   ALLOCATE (m(n_min), end_x(n_min), end_y(n_min))
   OPEN (UNIT=20, FILE=TRIM(file_min), STATUS='OLD', IOSTAT=err)
   IF (err /= 0) THEN
      WRITE (6, '(/,2A,/)') 'ERROR: Could not open ', TRIM(file_min)
      STOP
   END IF
   e_gmin = HUGE(e_gmin)
   DO i=1, n_min
      READ (20, *) j, m(i)
      IF (m(i) < e_gmin) THEN
         e_gmin = m(i)
         gmin = i
      END IF
   END DO
   CLOSE (20)
   WRITE (6, '(I6, 3A)') n_min, ' minima read in from ', TRIM(file_min), '.'
   IF ((lowest <= 0).OR.(lowest > n_min)) lowest = n_min


!  Procure transition state info.
   CALL count_ts(file_ts, n_ts, b_high, b_low)
   ALLOCATE (ts(n_ts))
   i = 1
   OPEN (UNIT=20, FILE=TRIM(file_ts), STATUS='OLD', IOSTAT=err)
   IF (err /= 0) THEN
      WRITE (6, '(/,2A,/)') 'ERROR: Could not open ', TRIM(file_ts)
      STOP
   END IF
   DO
      READ (20, *) ts(i)%n, ts(i)%e, lpafs, pg, h_pg, ts(i)%min1, ts(i)%min2
      IF (ts(i)%min1 /= ts(i)%min2) i=i+1
      IF (i == n_ts+1) EXIT
   END DO
   CLOSE (20)
   WRITE (6, '(I6, 3A)') &
      & n_ts, ' non-degenerate paths read in from ', TRIM(file_ts), '.'
   WRITE (6, '(A, 2(F18.10))') 'Highest and lowest transition states: ', &
      & b_high, b_low

   ALLOCATE (nodes(n_levels))
   ALLOCATE (basin(n_levels, n_min))
   WRITE (6, *)


!  Check existence of pick file rather than doing basin analysis and then failing.
   IF (file_pick /= '') THEN
      OPEN (UNIT=20, FILE=TRIM(file_pick), STATUS='OLD', IOSTAT=err)
      IF (err /= 0) THEN
         WRITE (6, '(/,2A,/)') 'ERROR: Could not open ', TRIM(file_pick)
         STOP
      END IF
      CLOSE (20)
   END IF


!  Reset transition state energies to the energy of the higher of the two minima
!  they connect if this option is turned on.
   IF (.NOT.barriers) THEN
      WRITE (6, '(A)') &
      'Reseting transition state energies to that of the higher connected minimum.'
      DO i = 1, n_ts
         ts(i)%e = MAX(m(ts(i)%min1), m(ts(i)%min2))
      END DO
      WRITE (6, '(A, /)') 'Done.'
   END IF


!  Assign the minima to their basins at each energy.
   WRITE (6, '(A)') 'Assigning minima to basins.'
   IF (split) THEN
      split_option = 0.0D0
   ELSE
      split_option = delta_e
   END IF
   basin = 0
   DO i = 1, n_levels
     energy = e_high - (i-1)*delta_e
     basin_no = 1
     DO j = 1, n_min
       IF ((basin(i, j) == 0).AND.(m(j) < energy+delta_e)) THEN
         basin(i, j) = basin_no
         DO
           again = .FALSE.
           DO k = 1, n_ts
             IF (ts(k)%e < energy+split_option) THEN
               min1 = ts(k)%min1
               min2 = ts(k)%min2
               IF (basin(i, min1) > basin(i, min2)) THEN
                 min_temp = min1
                 min1 = min2
                 min2 = min_temp
               ENDIF
               IF ((basin(i, min1)==0).AND.(basin(i, min2)==basin_no) &
                 & .AND.(m(min1) < energy+delta_e)) THEN
                 basin(i, min1) = basin_no
                 again = .TRUE.
               ENDIF
             ENDIF
           END DO
           IF (.NOT.again) EXIT
         END DO
         basin_no = basin_no + 1
       ENDIF
     END DO
     nodes(i) = basin_no - 1
     IF (nodes(i) == 1) THEN
        WRITE (6, '(I6, A, F18.10)') 1, ' basin  at energy ', energy
     ELSE
        WRITE (6, '(I6, A, F18.10)') nodes(i), ' basins at energy ', energy
     ENDIF
   END DO
   WRITE (6, '(A, /)') 'Done.'


!  Various pruning processes follow.
   WRITE (6, '(A)') 'Pruning minima.'

   basin = -basin

!  1. Remove nodes that don't lead to the lowest energy minima.
   ALLOCATE (indx1(n_min))
   CALL indexx(n_min, m, indx1)
   DO i = 1, lowest
      DO j = 1, n_levels
         basin(j, indx1(i)) = ABS(basin(j, indx1(i)))
      END DO
   END DO
   DEALLOCATE (indx1)

!  2. Remove nodes that are not monotonic sequence basin bottoms.
   IF (monotonic) THEN
      DO k = 1, n_ts
         IF (m(ts(k)%min1) < m(ts(k)%min2)) THEN
            min1 = ts(k)%min2
         ELSE
            min1 = ts(k)%min1
         END IF
         DO j = 1, n_levels
            basin(j, min1) = -ABS(basin(j, min1))
         END DO
      END DO
   END IF

!  3. Exclude all nodes.
   IF (excludeall) THEN
      basin = -ABS(basin)
   END IF

!  4. Deal with individual exceptions picked by user.
   IF (file_pick /= '') THEN
      OPEN (UNIT=20, FILE=TRIM(file_pick), STATUS='OLD', IOSTAT=err)
      IF (err /= 0) THEN
         WRITE (6, '(/,2A,/)') 'ERROR: Could not open ', TRIM(file_pick)
         STOP
      END IF
      DO
         READ (UNIT=20, FMT=*, IOSTAT=err) i
         IF (err /= 0) EXIT
         IF (ABS(i)==0 .OR. ABS(i)>n_min) THEN
            WRITE (6, '(A,I8,3A,I6)') 'WARNING: Ignoring ', i, ' in ', &
               TRIM(file_pick), '; minima numbers: 1 to', n_min
         ELSE
            IF (i < 0) THEN
               DO j = 1, n_levels
                  basin(j, -i) = -ABS(basin(j, -i))
               END DO
            ELSE
               DO j = 1, n_levels
                  basin(j, i) = ABS(basin(j, i))
               END DO
            END IF
         END IF
      END DO
      CLOSE (20)
   END IF

!  Renumber basins consecutively after pruning.
   ALLOCATE (indx1(0:MAXVAL(nodes)))
   DO i = 1, n_levels
      indx1 = 0
      k = 0
      DO j = 1, n_min
         p = basin(i, j)
         IF (p > 0) THEN
            IF (indx1(p) == 0) THEN
               k = k + 1
               indx1(p) = k
            END IF
         ELSE
            basin(i, j) = 0
         END IF
      END DO
      nodes(i) = k
      WRITE (6, '(A, I3, A, I6)') 'Nodes remaining at level ', i, ': ', k
      DO j = 1, n_min
         basin(i, j) = indx1(basin(i, j))
      END DO
   END DO
   DEALLOCATE (indx1)

   WRITE (6, '(A, /)') 'Done.'


!  Work out how many branches sprout from each node.

   WRITE (6, '(A)') 'Working out basin branching numbers.'
   ALLOCATE (branches(n_levels, MAXVAL(nodes)))
   branches = 0
   DO i = 1, n_levels-1
      k = 1
      DO j = 1, n_min
         IF (basin(i+1, j) == k) THEN
            branches(i, basin(i, j)) = branches(i, basin(i, j)) + 1
            k = k + 1
         ENDIF
      END DO
   END DO
   WRITE (6, '(A, /)') 'Done.'


!  Count the number of minima represented by each node.

   WRITE (6, '(A)') 'Counting node sizes.'
   ALLOCATE (node_size(n_levels, MAXVAL(nodes)))
   node_size = 0
   DO i = 1, n_levels
      DO j = 1, n_min
         k = basin(i, j)
         IF (k > 0) THEN
            node_size(i, k) = node_size(i, k) + 1
         END IF
      END DO
   END DO
   WRITE (6, '(A, /)') 'Done.'


!  Work out order in which to print the nodes in each level.

   WRITE (6, '(A)') 'Ordering branches.'
   ALLOCATE (order(n_levels, MAXVAL(nodes)))
   order = 0
   WRITE (6, '(A)') 'Level  1'
   DO i = 1, nodes(1)
      order(1, i) = i
   END DO
   DO i = 2, n_levels
      WRITE (6, '(A, I2)') 'Level ', i
      IF (order(i-1, 1) == 0) EXIT
      basin_gmin = basin(i, gmin)
      k = 1
      DO j = 1, nodes(i-1)
         n_br = branches(i-1, order(i-1, j))
         IF (n_br > 0) THEN
            ALLOCATE (connect(n_br), sorted(n_br), indx1(n_br))
            CALL connections(i-1, order(i-1, j), connect, n_br)
            DO p = 1, n_br
               sorted(p) = node_size(i, connect(p))
            END DO
            CALL indexx(n_br, DBLE(sorted), indx1)
            DO p = n_br, 1, -1
               IF (1.0D0*sorted(indx1(p))/sorted(indx1(n_br)) < 0.5D0) EXIT
            END DO
            big_ones = n_br-p
            CALL mix(n_br-big_ones, indx1(1 : n_br-big_ones))
            CALL mix(big_ones, indx1(n_br-big_ones+1 : n_br))
            padding = 1.0D0*(n_br - big_ones)/(big_ones + 1)
            DO p = 1, NINT(padding)
               order(i, k) = connect(indx1(p))
               k = k + 1
            END DO
            DO q = 1, big_ones
               order(i, k) = connect(indx1(n_br+1-q))
               k = k + 1
               DO p = NINT(padding*q)+1, NINT(padding*(q+1))
                  order(i, k) = connect(indx1(p))
                  k = k + 1
               END DO
            END DO
            IF (centre_gmin) THEN
               DO p = k-n_br, k-1
                  IF (order(i, p) == basin_gmin) THEN
                     q = k-n_br + INT(n_br/2.0D0)
                     s = order(i, q)
                     order(i, q) = basin_gmin
                     order(i, p) = s 
                     EXIT
                  END IF
               END DO
            END IF
            DEALLOCATE (connect, sorted, indx1)
         ENDIF
      END DO
   END DO
   WRITE (6, '(A, /)') 'Done.'


!  Dump the number of minima represented by each node in the order they will appear.

   IF (dump_sizes) THEN
      WRITE (6, '(A)') 'Writing node sizes to "node_sizes".'
      OPEN (UNIT=20, FILE='node_sizes', STATUS='REPLACE')
      DO i = 1, n_levels
         WRITE (20, '(A, I2, A, F20.10)') 'Level: ', i, ' energy ', e_high - (i-1)*delta_e
         DO j = 1, nodes(i)
            WRITE (20, '(I5, A, I7)') j, ':', node_size(i, order(i, j))
         END DO
         WRITE (20, *)
      END DO
      CLOSE (20)
      WRITE (6, '(A, /)') 'Done.'
   END IF


!  Dump the minima numbers at each node if required.
!  This double loop for each level is probably not the most efficient way, but...

   IF (dump_numbers) THEN
      WRITE (6, '(A)') &
      'Writing minima numbers associated with nodes to "node_numbers".'
      OPEN (UNIT=20, FILE='node_numbers', STATUS='REPLACE')
      DO i = 1, n_levels
         WRITE (20, '(/, A, I3, /, A)') 'LEVEL ', i, '========='
         DO j = 1, nodes(i)
            WRITE (20, '(/, A, I6)') 'Node ', j
            k = order(i, j)
            DO p = 1, n_min
               IF (basin(i, p) == k) WRITE (20, '(I6)') p
            END DO
         END DO
      END DO
      CLOSE (20)
      WRITE (6, '(A, /)') 'Done.'
   END IF


!  Find the parent of each basin.

   WRITE (6, '(A)') 'Finding the parent of each basin.'
   ALLOCATE (parent(n_levels, MAXVAL(nodes)))
   DO i = 2, n_levels
      DO j = 1, n_min
         IF (basin(i, j) > 0) THEN
            parent(i, basin(i, j)) = basin(i-1, j)
         END IF
      END DO
   END DO
   WRITE (6, '(A, /)') 'Done.'


!  Allocate virtual columns to each node.

   WRITE (6, '(A)') 'Allocating vertical space to nodes.'

!  1. Number of columns for graph = total number of minima plotted.
   n_cols = 0
   DO i = 1, nodes(1)
      n_cols = n_cols + node_size(1, i)
   END DO

!  2.  Allocate columns to first-row nodes.
   ALLOCATE (col_0(n_levels, MAXVAL(nodes)))
   j = 0
   DO i = 1, nodes(1)
      col_0(1, order(1, i)) = j + 1
      j = j + node_size(1, order(1, i))
   END DO

!  3. Allocate to each node in successive levels a fraction of the vertical space
!     allocated to its parent.
   ALLOCATE (indx1(MAXVAL(nodes)))
   DO i = 2, n_levels
      indx1 = col_0(i-1, :)
      DO j = 1, nodes(i)
         k = order(i, j)
         col_0(i, k) = indx1(parent(i, k))
         indx1(parent(i, k)) = indx1(parent(i, k)) + node_size(i, k)
      END DO
   END DO
   DEALLOCATE (indx1)

   WRITE (6, '(A, /)') 'Done.'


!  Generate the PostScript file.

   Write (6, '(A)') 'Writing tree.ps.'
   end_x = 0.0D0
   end_y = 0.0D0
   OPEN (UNIT=20, FILE='tree.ps', STATUS='REPLACE')
   CALL header
   DO i = 2, n_levels
      WRITE (20, '(/, A, I2, A, I2, /)') '% Level: ', i-1, ' to ', i
      IF (nodes(i-1) == 0) EXIT
      energy = e_high - (i-2)*delta_e
      y1 = y_pos(1.0D0*(i-2)/(n_levels-1))
      y2 = y_pos(1.0D0*(i-1)/(n_levels-1))
      DO j = 1, nodes(i)
         k = order(i, j)
         x1 = (col_0(i-1, parent(i,k)) + node_size(i-1, parent(i,k))/2.0 - 1.0) / n_cols
         x2 = (col_0(i, k) + node_size(i, k)/2.0 - 1.0) / n_cols
         x1 = x_pos(x1)
         x2 = x_pos(x2)
         IF (branches(i,k) == 0) THEN
            CALL endpoint(i, k, end_e, end_m)
            frac = (energy-end_e)/delta_e
            y3 = frac*(y2-y1) + y1
            x3 = frac*(x2-x1) + x1
            IF (y3 < y2) y3=y2
            end_x(end_m) = x3
            end_y(end_m) = y3
            WRITE (20, '(2F7.2, A, 2F7.2, A)') x1, y1, ' mt ', x3, y3, ' ls'
         ELSE
            WRITE (20, '(2F7.2, A, 2F7.2, A)') x1, y1, ' mt ', x2, y2, ' ls'
         END IF
      END DO
   END DO
   IF (identify) THEN
      WRITE (20, '(A, I2, A)') &
         & '/Times-Roman findfont ', label_size, ' scalefont setfont'
      DO i = 1, n_min
         IF ( (end_x(i) > 0.0D0).AND.(end_y(i) > 0.0D0) ) THEN
            WRITE (min_trim, '(I8)') i
            WRITE (20, '(2F7.2, 3A)') end_x(i), end_y(i), &
               & ' mt (', TRIM(ADJUSTL(min_trim)), ') show'
         END IF
      END DO
   END IF
   WRITE (20, '(/, A)') 'showpage'
   CLOSE (20)
   WRITE (6, '(A, /)') 'Done.'

END PROGRAM disconnection

!................................................................................!

! Read options from the file dinfo.

SUBROUTINE read_options

   USE keywords
   USE page
   USE vars
   IMPLICIT NONE
   CHARACTER(LEN=50) :: keyword
   INTEGER err
   LOGICAL :: success

   barriers = .TRUE.
   centre_gmin = .FALSE.
   delta_e = 0.0D0
   dump_numbers = .FALSE.
   dump_sizes = .FALSE.
   e_high = 0.0D0
   excludeall = .FALSE.
   lowest = 0
   lab_fmt = 'F6.1'
   monotonic = .FALSE.
   n_levels = 0
   file_pick = ''
   split = .TRUE.

   OPEN (UNIT=20, FILE='dinfo', STATUS='OLD', IOSTAT=err)
   IF (err /= 0) THEN
      WRITE (6, '(/,A,/)') 'ERROR: Could not open dinfo file'
      STOP
   END IF
   DO
      CALL read_line(20, success)
      IF (.NOT.success) EXIT
      CALL get_string(keyword)
      CALL upper_case(keyword)
      IF (keyword(1:1) == '#') keyword='#'
      SELECT CASE(TRIM(keyword))
      CASE ('#')
!        Comment; do nothing.
      CASE ('CENTREGMIN')
         centre_gmin = .TRUE.
      CASE ('DELTA')
         CALL get_dp(delta_e)
      CASE ('DUMPNUMBERS')
         dump_numbers = .TRUE.
      CASE ('DUMPSIZES')
         dump_sizes = .TRUE.
      CASE ('EXCLUDEALL')
         excludeall = .TRUE.
      CASE ('FIRST')
         CALL get_dp(e_high)
      CASE ('IDENTIFY')
         identify = .TRUE.
      CASE ('LABELFORMAT')
         CALL get_string(lab_fmt)
      CASE ('LETTER')
         page_x = 612
         page_y = 792
      CASE ('LEVELS')
         CALL get_integer(n_levels)
      CASE ('LOWEST')
         CALL get_integer(lowest)
      CASE ('MINIMA')
         CALL get_string(file_min)
      CASE ('MONOTONIC')
         monotonic = .TRUE.
      CASE ('NOBARRIERS')
         barriers = .FALSE.
      CASE ('NOSPLIT')
         split = .FALSE.
      CASE ('PICK')
         CALL get_string(file_pick)
      CASE ('TS')
         CALL get_string(file_ts)
      CASE DEFAULT
         WRITE (6, '(2A)') 'Keyword not recognised in dinfo: ', TRIM(keyword)
         STOP
      END SELECT
   END DO
   CLOSE (20)

   IF (delta_e <= 0.0D0) THEN
      WRITE (6, '(A)') 'Energy gap must be specified using DELTA.'
      STOP
   ENDIF

   IF (e_high == 0.0D0) THEN
      WRITE (6, '(A)') 'First energy level must by specified using FIRST.'
      STOP
   ENDIF

   IF (n_levels <= 0) THEN
      WRITE (6, '(A)') 'Number of levels must be specified using LEVELS.'
      STOP
   ENDIF

END SUBROUTINE read_options

!................................................................................!
 
! Count the number of minima in the list.
 
SUBROUTINE count_min(fn, n)
 
   IMPLICIT NONE
 
   CHARACTER(LEN=100), INTENT(IN) :: fn
   INTEGER, INTENT(OUT) :: n
 
   INTEGER :: err, j
 
   n = 0
   OPEN (UNIT=20, FILE=TRIM(fn), STATUS='OLD', IOSTAT=err)
   IF (err /= 0) THEN
      WRITE (6, '(/,2A,/)') 'ERROR: Could not open ', TRIM(fn)
      STOP
   END IF
   DO
      READ (UNIT=20, FMT=*, IOSTAT=err) j
      IF (err /= 0) EXIT
      n = n + 1
   END DO
   CLOSE (20)
 
END SUBROUTINE count_min

!................................................................................!

! Count the number of transition states for NON-DEGENERATE rearrangements.
! Also returned are the highest and lowest transition state energies.

SUBROUTINE count_ts(fn, n_ts, b_high, b_low)

   IMPLICIT NONE

   CHARACTER(LEN=100), INTENT(IN) :: fn
   INTEGER, INTENT(OUT) :: n_ts
   DOUBLE PRECISION :: b_high, b_low

   CHARACTER(LEN=3) :: pg
   DOUBLE PRECISION :: e, lpafs
   INTEGER :: err, i, h, min1, min2

   n_ts = 0
   b_high = -HUGE(b_high)
   b_low = HUGE(b_high)
   OPEN (UNIT=20, FILE=TRIM(fn), STATUS='OLD', IOSTAT=err)
   IF (err /= 0) THEN
      WRITE (6, '(/,2A,/)') 'ERROR: Could not open ', TRIM(fn)
      STOP
   END IF
   DO
      READ (UNIT=20, FMT=*, IOSTAT=err) i, e, lpafs, pg, h, min1, min2
      IF (err /= 0) EXIT
      IF (min1 /= min2) n_ts=n_ts+1
      IF (e > b_high) b_high=e
      IF (e < b_low) b_low=e
   END DO
   CLOSE (20)

END SUBROUTINE count_ts

!................................................................................!

! Subroutine to make a list (CONNECT) of the basins in level LEVEL+1 that are
! connected to basin NODE in level LEVEL.

SUBROUTINE connections(level, node, connect, n_br)

   USE vars
   IMPLICIT NONE

   INTEGER, INTENT(IN) :: level, node, n_br
   INTEGER, DIMENSION(n_br), INTENT(OUT) :: connect

   INTEGER :: i, j, last

   last = 0
   j = 1
   connect = 0
   DO i = 1, n_min
      IF ((basin(level, i) == node).AND.(basin(level+1, i) > last)) THEN
         connect(j) = basin(level+1, i)
         last = connect(j)
         j = j + 1
      ENDIF
   END DO

END SUBROUTINE connections

!................................................................................!

! Write preamble for PostScript file and draw energy scale.

SUBROUTINE header

   USE page
   USE vars
   IMPLICIT NONE

   DOUBLE PRECISION :: x1, x2, y1, y2, x_pos, y_pos
   DOUBLE PRECISION, PARAMETER :: tick_width=3.0D0
   INTEGER :: i

!  Definitions.
   WRITE (20, '(6(A, /), A)') '%!PS', &
      & '%% Creator: disconnection.f90', &
      & '%% Minima: ' // TRIM(file_min), &
      & '%% Transition states: ' // TRIM(file_ts), &
      & '/ls {lineto stroke} def', &
      & '/mt {moveto} def', &
      & '0.1 setlinewidth'

!  Vertical energy axis.
   x1 = margin_x + scale_width - 10.0D0
   y1 = y_pos(0.0D0)
   y2 = y_pos(1.0D0)
   WRITE (20, '(2F7.2, A, 2F7.2, A)') x1, y1, ' mt ', x1, y2, ' ls'

!  Ticks on axis.
   x2 = x1 + tick_width/2.0D0
   x1 = x1 - tick_width/2.0D0
   DO i=1, n_levels
      y1 = y_pos(1.0D0*(i-1)/(n_levels-1))
      WRITE (20, '(2F7.2, A, 2F7.2, A)') x1, y1, ' mt ', x2, y1, ' ls'
   END DO

!  Tick labels.
   WRITE (20, '(A, I2, A)') &
      & '/Times-Roman findfont ', font_size, ' scalefont setfont'
   DO i=1, n_levels
      x1 = 15.0
      y1 = y_pos(1.0D0*(n_levels-i)/(n_levels-1)) - 5.0D0
      WRITE (20, '(2F7.2, A, '//TRIM(lab_fmt)//', A)') &
         & x1, y1, ' mt (', e_high - (n_levels-i)*delta_e, ') show'
   END DO

!  Landmark for manipulate.f90.
   WRITE (20, '(/, A, /)') '% Manipulate landmark.'

END SUBROUTINE header

!................................................................................!

! Convert coordinates on (0,1) to PostScript units on paper.

FUNCTION x_pos(x)

   USE page
   IMPLICIT NONE
   DOUBLE PRECISION :: x_pos
   DOUBLE PRECISION, INTENT(IN) :: x

   x_pos = x * (page_x - 2*margin_x - scale_width) + margin_x + scale_width

END FUNCTION x_pos

FUNCTION y_pos(y)

   USE page
   IMPLICIT NONE
   DOUBLE PRECISION :: y_pos
   DOUBLE PRECISION, INTENT(IN) :: y
 
   y_pos = page_y - (y * (page_y - 2*margin_y) + margin_y)
 
END FUNCTION y_pos

!................................................................................!

SUBROUTINE endpoint(l, b, end_e, end_m)

   USE vars
   IMPLICIT NONE

   INTEGER, INTENT(IN) :: l, b
   DOUBLE PRECISION, INTENT(OUT) :: end_e
   INTEGER, INTENT(OUT) :: end_m

   INTEGER :: a

   end_e = HUGE(end_e)

   DO a = 1, n_min
      IF ((basin(l, a) == b).AND.(m(a) < end_e)) THEN
         end_e = m(a)
         end_m = a
      END IF
   END DO

END SUBROUTINE endpoint

!................................................................................!

SUBROUTINE mix(n, mixlist)

   IMPLICIT NONE

   INTEGER :: n
   INTEGER, DIMENSION(n) :: mixlist

   INTEGER :: sgn, delta, pos, i
   INTEGER, DIMENSION(n) :: copylist
   LOGICAL :: big

   copylist = mixlist
   pos = 1
   sgn = 1
   delta = n-1
   big = .TRUE.

   DO i = 1, n
      mixlist(i) = copylist(pos)
      IF (big) THEN
         pos = pos + sgn*delta
         delta = delta - 2
         sgn = -sgn
         big = .FALSE.
      ELSE
         pos = pos + sgn
         big = .TRUE.
      END IF
   END DO

END SUBROUTINE mix

!................................................................................!

! Unmodified Numerical Recipes routine to return a sorted index indx of
! length n for the array arr such that arr(inx(i)) are in ascending order
! for i=1, 2, ... n.

      SUBROUTINE indexx(n,arr,indx)
      INTEGER n,indx(n),M,NSTACK
      DOUBLE PRECISION arr(n)
      PARAMETER (M=7,NSTACK=50)
      INTEGER i,indxt,ir,itemp,j,jstack,k,l,istack(NSTACK)
      DOUBLE PRECISION a
      do 11 j=1,n
        indx(j)=j
11    continue
      jstack=0
      l=1
      ir=n
1     if(ir-l.lt.M)then
        do 13 j=l+1,ir
          indxt=indx(j)
          a=arr(indxt)
          do 12 i=j-1,1,-1
            if(arr(indx(i)).le.a)goto 2
            indx(i+1)=indx(i)
12        continue
          i=0
2         indx(i+1)=indxt
13      continue
        if(jstack.eq.0)return
        ir=istack(jstack)
        l=istack(jstack-1)
        jstack=jstack-2
      else
        k=(l+ir)/2
        itemp=indx(k)
        indx(k)=indx(l+1)
        indx(l+1)=itemp
        if(arr(indx(l+1)).gt.arr(indx(ir)))then
          itemp=indx(l+1)
          indx(l+1)=indx(ir)
          indx(ir)=itemp
        endif
        if(arr(indx(l)).gt.arr(indx(ir)))then
          itemp=indx(l)
          indx(l)=indx(ir)
          indx(ir)=itemp
        endif
        if(arr(indx(l+1)).gt.arr(indx(l)))then
          itemp=indx(l+1)
          indx(l+1)=indx(l)
          indx(l)=itemp
        endif
        i=l+1
        j=ir
        indxt=indx(l)
        a=arr(indxt)
3       continue
          i=i+1
        if(arr(indx(i)).lt.a)goto 3
4       continue
          j=j-1
        if(arr(indx(j)).gt.a)goto 4
        if(j.lt.i)goto 5
        itemp=indx(i)
        indx(i)=indx(j)
        indx(j)=itemp
        goto 3
5       indx(l)=indx(j)
        indx(j)=indxt
        jstack=jstack+2
        if(jstack.gt.NSTACK)pause 'NSTACK too small in indexx'
        if(ir-i+1.ge.j-l)then
          istack(jstack)=ir
          istack(jstack-1)=i
          ir=j-1
        else
          istack(jstack)=j-1
          istack(jstack-1)=l
          l=i
        endif
      endif
      goto 1
      END
!  (C) Copr. 1986-92 Numerical Recipes Software 1(-V%'2150)-3.
